package things;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonView;
import com.fasterxml.jackson.core.JsonProcessingException;

import java.util.Collection;

/**
 * Represents a thing. A thing can contain multiple inputs and outputs.
 *
 * @see Resource
 */
public class Thing extends NamedResource {
    private final ResourceList<IO> ios = new ResourceList<>();

    Thing(String name) {
        super(name);
    }

    /**
     * Returns the collection of all input and output resources of the thing.
     *
     * @return  All inputs/outputs.
     */
    @JsonProperty("io")
    @JsonView(JSONViews.ThingDetail.class)
    public Collection<IO> getIOs() {
        return ios;
    }

    /**
     * Returns the input or output with the given name or null if no input/output with the given name exists.
     *
     * @param name  Name of the input/output.
     * @return      Input/output or null if no object with the given name exists.
     */
    public IO getIOWithName(String name) {
        return ios.find(name);
    }

    /**
     * Adds an input with the given name.
     * If the input could be added, the newly created input is returned. If an object with the given name already exists within the thing, null is returned.
     *
     * @param name  Name of the input to add.
     * @return      Newly added input, null if input could not be added.
     */
    public Input addInputWithName(String name) {
        Input input = new Input(this, name);
        if (ios.add(input)) {
            return input;
        } else {
            return null;
        }
    }

    /**
     * Adds an output with the given name.
     * If the output could be added, the newly created output is returned. If an object with the given name already exists within the thing, null is returned.
     *
     * @param name  Name of the output to add.
     * @return      Newly added output, null if input could not be added.
     */
    public Output addOutputWithName(String name) {
        Output output = new Output(this, name);
        if (ios.add(output)) {
            return output;
        } else {
            return null;
        }
    }

    ResourceList<IO> getIOResources() {
        return ios;
    }

    @Override
    public Type getType() {
        return Type.THING;
    }

    @Override
    public String getURI() {
        return "/things/" + getName();
    }

    @Override
    public String getJSON() throws JsonProcessingException {
        return Things.sharedObjectMapper().writerWithView(JSONViews.ThingDetail.class).writeValueAsString(this);
    }
}
