package things;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;

import java.util.Arrays;
import java.util.Collection;

/**
 * Allows to store a hierarchy of resources.
 */
public class Things extends NamedResource {
    private final static ObjectMapper objectMapper = new ObjectMapper();

    static ObjectMapper sharedObjectMapper() {
        return objectMapper;
    }

    private final ResourceList<Thing> things = new ResourceList<>();

    /**
     * Creates a new Things instance that can be used to save Thing, Input and Output resources.
     *
     * Note that this is the only class your servlet has to interact directly with. It has all the functions needed to create, find, modify and delete all kind of resources.
     */
    public Things() {
        super("things");
        objectMapper.enable(SerializationFeature.INDENT_OUTPUT);
    }

    /**
     * Returns the collection of all things contained in this collection.
     *
     * @return  All Things part of the collection.
     */
    public Collection<Thing> getThings() {
        return things;
    }

    /**
     * Returns the thing with the given name or null if no such thing exists.
     *
     * @param name  Name of the thing to look for.
     * @return      The Thing instance with the given name or null if no such thing exists.
     */
    public Thing getThingWithName(String name) {
        return things.find(name);
    }

    /**
     * Tries to add a new Thing instance with the given name and returns the instance just added. If there is already a thing with the same name, no new thing will be added and null is returned
     * instead.
     *
     * @param name  Name of the thing to add.
     * @return      The newly created thing or null if a thing with the same name already exists.
     */
    public Thing addThingWithName(String name) {
        Thing thing = new Thing(name);
        if (things.add(thing)) {
            return thing;
        } else {
            return null;
        }
    }

    /**
     * Tries to find the resource identified by the given URI. Returns the resource or null if the resource could not be found.
     *
     * @param uri   URI of the resource to look for.
     * @return      Resource instance or null if no resource exists at the given URI.
     */
    public Resource getResourceByURI(String uri) {
        String[] components = parseURI(uri);

        if (components.length == 0 || !components[0].equals("things")) {
            return null;
        }

        switch (components.length) {
            case 1:
                return this;

            case 2:
                return things.find(components[1]);

            case 3:
                Thing thing = things.find(components[1]);
                if (thing == null) {
                    return null;
                }
                return thing.getIOResources().find(components[2]);

            default:
                return null;
        }
    }

    /**
     * Creates a new child resource to the parent resource at the given  URI of the given type. Returns the created resource or null if the resource could not be created.
     *
     * @param parentResourceURI URI of the parent resource to which the new resource should be added.
     * @param childName         Name for the new child resource.
     * @param childType         Type of the child resource to create.
     * @return                  The child resource created or null if the resource could not be created.
     */
    public Resource addChildToResourceWithURI(String parentResourceURI, String childName, Resource.Type childType) {

        // Get the root resource to which the item should be added.
        Resource resource = getResourceByURI(parentResourceURI);

        // Ensure that this resource exists.
        if (resource == null) {
            return null;
        }

        // For each different type of parent resource, there is another action to take.
        switch (resource.getType()) {
            case THINGS:

                // Try to add a new thing.
                return addThingWithName(childName);

            case THING:
                Thing thing = (Thing)resource;

                // Check what type of resource has to be added to the thing.
                switch (childType) {
                    case INPUT:

                        // Try to add a new input.
                        return thing.addInputWithName(childName);

                    case OUTPUT:

                        // Try to add a new output.
                        return thing.addOutputWithName(childName);

                    default:

                        // Fail if unsupported type of object has to be created.
                        return null;
                }

            case INPUT:
            case OUTPUT:

                // Fail, input and outputs do not have child resources.
                return null;

            default:

                // If the parent resource is of unknown type, fail.
                return null;
        }
    }

    /**
     * Removes the resource (whatever type it is) from the resource model. All it's children will be removed too.
     *
     * @param uri   URI of the resource to remove.
     * @return      The resource if it exists and could be removed, null otherwise.
     */
    public Resource removeResourceByURI(String uri) {
        Resource resource = getResourceByURI(uri);

        if (resource == null) {
            return null;
        }

        switch (resource.getType()) {
            case THING:
                things.remove(resource);
                return resource;

            case INPUT:
            case OUTPUT:
                ((IO)resource).getParent().getIOResources().remove(resource);
                return resource;

            default:
                return null;
        }
    }

    /**
     * Adds sample resources to the collection.
     */
    public void addSampleResources() {
        things.clear();
        addChildToResourceWithURI("/things", "thing1", Type.THING);
        addChildToResourceWithURI("/things/thing1", "input1", Type.INPUT);
        addChildToResourceWithURI("/things/thing1", "output1", Type.OUTPUT);
        addChildToResourceWithURI("/things", "thing2", Type.THING);
        addChildToResourceWithURI("/things/thing2", "input1", Type.INPUT);
        addChildToResourceWithURI("/things/thing2", "output1", Type.OUTPUT);
    }

    @Override
    public Type getType() {
        return Type.THINGS;
    }

    @Override
    public String getURI() {
        return "/things";
    }

    @Override
    public String getJSON() throws JsonProcessingException {
        return objectMapper.writerWithView(JSONViews.ThingOverview.class).writeValueAsString(things);
    }

    private String[] parseURI(String uri) {
        String[] components = uri.split("/", 0);
        if (components.length > 0 && components[0].isEmpty()) {
            components = Arrays.copyOfRange(components, 1, components.length);
        }
        return components;
    }
}
